CACHE_MINUTE = 60
CACHE_15_MINS = 900
CACHE_30_MINS = 1800
CACHE_HOUR = 3600
CACHE_8_HOURS = 28800
CACHE_12_HOURS = 43200
CACHE_DAY = 86400
CACHE_2_DAYS = 172800
CACHE_WEEK = 604800
CACHE_2_WEEKS = 1209600
CACHE_MONTH = 2592000



from cache_money import utils as cache_utils
from cache_money.cache import cache_money, close_cache_money, init_cache_money

__all__ = ["cache_money", "close_cache_money", "init_cache_money", "cache_utils"]



import hashlib
import inspect
import logging
import pickle
from functools import wraps
from typing import Any, Callable, List, Optional, Union

import aioredis

from cache_money import engine
from cache_money.constants import CACHE_HOUR
from cache_money.utils import decode, get_function_path

log = logging.getLogger(__name__)


class CacheMoney(object):
    """
    Walrus-inspired redis cache implementation that supports asyncio. Implement simple GET/SET/DELETE operations,
    and a decorator.
    """

    def __init__(self):
        self.conn: Optional[aioredis.client.Redis] = None
        self.prefix: Optional[str] = None
        self.default_timeout: Optional[int] = None
        self.enabled: bool = False

    def setup_cache(
        self,
        conn: Optional[aioredis.client.Redis] = None,
        prefix: Optional[str] = None,
        default_timeout: Optional[int] = None,
        enabled: bool = False,
    ):
        """
        Setup the caching.

        Args:
            conn: Active aioredis client, optional when in debug.
            prefix: A prefix added in front of the cache key.
            default_timeout: Default timeout for all cache entries that don't specify their own time.
            enabled: Whether the cache is active or not.
        """
        self.conn = conn
        self.prefix = prefix
        self.default_timeout = default_timeout
        self.enabled = enabled

    def make_key(self, keys: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Apply the prefix in front of the key separated by a colon `:` if the prefix has been set for the class.

        Args:
            keys: Keys to add the prefix to.

        Returns:
            Keys with the prefix added.BBB
        """
        if isinstance(keys, str):
            return ":".join((self.prefix, keys)) if self.prefix else keys
        return [(":".join((self.prefix, k)) if self.prefix else k) for k in keys]

    async def get(self, key: str, default: Any = None, transform_key: bool = True) -> Any:
        """
        Retrieve a value from the cache. In the event the value does not exist, return the `default`.

        Trying to access a non-existing key in Redis returns None from aioredis. This effectively makes `None` a non
        cacheable value for Cache Money.

        Params:
            key: The redis key to get.
            default: The default value to return in case that the value cannot be retrieved.
            transform_key: Whether to apply the transformation function to the key (make_key).

        Returns:
            The value stored in redis (de-pickled) or the default value.
        """
        if not self.enabled:
            return default

        if transform_key:
            key = self.make_key(key)

        try:
            value = await self.conn.get(key)
        except Exception:
            log.exception(f"Exception trying to retrieve key [{key}] from Redis")
            return default

        if value is not None:
            try:
                return pickle.loads(value)
            except Exception:
                log.exception(f"Exception trying to load the pickled object from key [{key}]")
                return default
        else:
            return value

    async def set(self, key: str, value: Any, timeout: int = None, transform_key: bool = True) -> bool:
        """
        Cache the given `value` in the specified `key`. If no timeout is specified, the default timeout will be used.

        Params:
            key: The redis key to set.
            value: The raw value to set (will be pickled).
            timeout: The timeout to set in Redis (in seconds)
            transform_key: Whether to apply the transformation function to the key (make_key).

        Returns:
            True if the set was successful, False otherwise.
        """
        if not self.enabled:
            return True

        if transform_key:
            key = self.make_key(key)

        if timeout is None:
            timeout = self.default_timeout

        try:
            pickled_value = pickle.dumps(value)
        except Exception:
            log.exception(f"Error pickling object for caching: key=[{key}]; value type=[{type(value)}]")
            return False

        try:
            if timeout:
                return await self.conn.setex(key, int(timeout), pickled_value)
            else:
                return await self.conn.set(key, pickled_value)
        except Exception:
            log.exception(f"Exception trying to set key [{key}] in Redis")
            return False

    async def delete(self, keys: Union[str, List[str], bytes, List[bytes]], make_key: bool = True):
        """
        Remove one or more keys from the cache.

        Args:
            keys: One or more keys to delete, can be string or byte string. You cannot mix and match strings and byte
                strings if you send a list.
            make_key: Whether the key construction function is to be ran on the keys received. If byte strings are
                received, the make_key function will not be executed.
        """
        if not self.enabled or not keys:
            return

        if make_key and isinstance(keys[0], str):
            keys = self.make_key(keys)

        keys = [keys] if not isinstance(keys, list) else keys

        return await self.conn.delete(*keys)

    async def bust(self, force: bool = False):
        """
        Delete all keys in Redis starting with the value for the prefix. If there are no value defined for the
        prefix, that means that there is no way to know which keys belongs to cache-money and which keys belong to
        something else using the redis DB Cache Money is connected to. If you want to delete keys without a prefix set
        for Cache Money, you need to pass force=True.
        """
        if not self.enabled:
            return

        if not self.prefix and not force:
            log.warning(
                "Cache Money skipped bust on all Cache Money values because no prefix has been set for "
                "Cache Money which will result in all keys on the redis DB being removed. You can pass "
                "force=True to the function `bust` to execute the bust without a prefix."
            )
            return

        keys = await self.keys(decode_to_str=False)
        return await self.delete(keys)

    async def keys(self, function: Callable = None, decode_to_str: bool = True) -> Union[List[bytes], List[str]]:
        """
        Return keys from Redis that were cached by Cache Money. Note that if the parameter function is not provided and
        Cache Money doesn't have a value for its prefix attribute, all keys in the Redis DB will be returned, even if
        the keys were not set up by Cache Money.

        Args:
            function: If provided will only returns keys representing cache entries for the function.
            decode_to_str: Whether the keys should be decoded before being returned to this function. If set to False
                the function will return byte strings instead of strings.

        Returns:
            A list of string or byte strings representing the cache keys in Redis.
        """
        if not self.enabled:
            return []

        function_path = get_function_path(function) if function else ""

        keys = await self.conn.keys(self.make_key(function_path) + "*")
        if decode_to_str:
            keys = [decode(key) for key in keys]
        return keys

    def _key_fn(a, k) -> str:
        """Generate a hash from args and kwargs."""
        return hashlib.md5(pickle.dumps((a, k))).hexdigest()

    def cached(self, key_fn: Callable = _key_fn, timeout: int = None) -> Any:
        """
        Decorator that will transparently cache calls to the wrapped function. By default, the cache key will be made
        up of the arguments passed in (like memoize), but you can override this by specifying a custom `key_fn`.

        Usage:
            from cache_money import cache_money
            from cache_money.constants import CACHE_MINUTE

            @cache.cached(timeout=CACHE_MINUTE)
            async def addition(x, y):
                return x + y

            await addition(3, 4)  # Function is called.
            await addition(3, 4)  # Not called, value is cached.
            await add_numbers.bust(3, 4)  # Clear cache for (3, 4).
            await add_numbers(3, 4)  # Function is called.

        The decorated function also gains a new attribute named `bust` which will clear the cache for the given args.

        Args:
            key_fn: Function used to generate a key from the given args and kwargs.
            timeout: Time to cache return values in seconds.

        Returns:
            Return the result of the decorated function call with the given args and kwargs.
        """

        def decorator(fn):
            def make_key(args, kwargs):
                """Apply the module and name of the function in front of the key, separated by a colon `:`."""
                return f"{inspect.getmodule(fn).__name__}.{fn.__name__}:{key_fn(args, kwargs)}"

            async def bust(*args, **kwargs):
                """Bust a specific cache entry for the decorated function that match the provided args and kwargs."""
                return await self.delete(make_key(args, kwargs))

            async def bust_all():
                """Bust all cache entries for the decorated function."""
                keys = await self.keys(fn, decode_to_str=False)
                return await self.delete(keys)

            @wraps(fn)
            async def inner(*args, **kwargs):
                if not self.enabled:
                    return await fn(*args, **kwargs)

                key = make_key(args, kwargs)
                res = await self.get(key)
                if res is None:
                    res = await fn(*args, **kwargs)

                    if res is not None:  # None is not cacheable in Cache Money
                        await self.set(key, res, timeout)

                return res

            inner.bust = bust
            inner.bust_all = bust_all
            inner.make_key = make_key
            return inner

        return decorator


cache_money: CacheMoney = CacheMoney()


def init_cache_money(
    host: str,
    port: int = 6379,
    prefix: str = None,
    db: int = 0,
    default_timeout: Optional[int] = CACHE_HOUR,
    enabled: bool = True,
):
    """
    Create the connection to Redis and setup Cache Money to use it.

    Args:
        host: redis host
        port: redis port
        prefix: An optional prefix to add in front of all cache keys generated.
        db: The database number in redis.
        default_timeout: A default timeout in seconds to set on function cache if no timeout is provided when
            decorating the function.
        enabled: If the cache is enabled or not. Set it to False when you don't want to enable cache for local
            development or for debugging. When set to False no connection to Redis will be established and all call to
            Cache Money will be ignored.
    """
    if not enabled:
        log.info("Initialization of Cache Money skipped because enabled=False")
        return

    global cache_money

    log.info("Initializing Cache Money..")

    conn = engine.get_connection()
    if not conn:
        conn = engine.init_connection(
            host=host,
            port=port,
            db=db,
        )

    cache_money.setup_cache(conn=conn, prefix=prefix, default_timeout=default_timeout, enabled=enabled)


async def close_cache_money():
    await engine.close_connection()



import logging
from typing import Optional

import aioredis

redis_conn: Optional[aioredis.client.Redis] = None

log = logging.getLogger(__name__)


def init_connection(host: str, port: int = 6379, db: int = 0) -> aioredis.client.Redis:
    global redis_conn

    url = f"redis://{host}:{port}"

    log.info(f"Connecting to redis {url}")
    redis_conn = aioredis.from_url(
        url=url,
        encoding="utf-8",
        db=db,
    )

    return redis_conn


def get_connection() -> aioredis.client.Redis:
    global redis_conn
    return redis_conn


async def close_connection():
    global redis_conn
    if redis_conn:
        await redis_conn.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cache_money']

package_data = \
{'': ['*']}

install_requires = \
['aioredis>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'cache-money',
    'version': '1.1.0',
    'description': 'Async cache decorator for memoization using aioredis.',
    'long_description': '# Cache Money\n\n[![Build Status](https://dev.azure.com/novisto/novisto/_apis/build/status/novisto.cache-money?branchName=master)](https://dev.azure.com/novisto/novisto/_build/latest?definitionId=30&branchName=master)\n[![codecov](https://codecov.io/gh/novisto/cache-money/branch/master/graph/badge.svg?token=V05Y6MTMU2)](https://codecov.io/gh/novisto/cache-money)\n[![Version](https://img.shields.io/pypi/v/cache-money)](https://pypi.org/project/cache-money/)\n[![Python Version](https://img.shields.io/pypi/pyversions/cache-money)](https://pypi.org/project/cache-money/)\n\nAsync cache library for [memoization](https://en.wikipedia.org/wiki/Memoization) using Redis. Inspired by \n[Walrus](https://github.com/coleifer/walrus) and implemented with [aioredis](https://github.com/aio-libs/aioredis-py).\n\nCache Money is used through a decorator you can add to your function that needs to be cached. When the decorator \ngets executed, Cache Money will make a unique key from the name of the function and the params received and look up in \nredis if there is a result for this key. If there is a result it will be used as the output of the function and the \nexecution of the function will be skipped.\n\nYou can add a timeout in the declaration of the decorator, you can find constants for common timeout duration in \n`cache_money/constants.py`. When the timeout is reached, Redis remove the entry itself.\n\nIt\'s also possible to clear the cache early by using the method bust that gets added to a function decorated by \nCache Money. An example is provided below.\n\nThis library is available on PyPI under the name cache-money. You can install with pip by running `pip install \ncache-money`.\n\n\n# Requirements\n\nYou need a redis instance running to use this library. This library was tested to run on version of Redis >= 4.0.0. \nIf you have docker set up you can create a redis instance like this:\n\n```shell\nmake redis-start\n```\n\n\n# Usage\n\n## Basic usage\n\nFirst thing is initializing Cache Money and decorating a function that you want to cache\n\n```python\nfrom cache_money import cache_money, init_cache_money\nfrom cache_money.constants import CACHE_HOUR, CACHE_WEEK\n\ninit_cache_money(host="localhost")\n\n@cache_money.cached(timeout=CACHE_HOUR)\nasync def addition(x: int, y: int) -> int:\n    return x + y\n\n@cache_money.cached(timeout=CACHE_WEEK)\nasync def multiplication(x: int, y: int) -> int:\n    return x * y\n```\n\nIf you run the following calls to the function `addition` consecutively:\n```doctest\n  >>> await addition(3, 4)\n  7\n  \n  >>> await addition(3, 7)\n  10\n \n  >>> await addition(3, 4)\n  7\n```\n\nThe first and second call would be executed, but the third call would have used the cache in redis instead, as long \nas the third call was done within one hour of when the first call was made, as the function addition is caching results \nfor one hour.\n\nIn Redis you would see two entries like this:\n\n```shell\n# redis-cli \n\n127.0.0.1:6379> KEYS *\n1) "__main__:addition:ea53056bad64a599c84efdfd4f4cbb64"\n2) "__main__:addition:bb6b7afb6a6cf3191f6d7fd35d976d42"\n\n127.0.0.1:6379> TTL addition:ea53056bad64a599c84efdfd4f4cbb64\n(integer) 3403\n```\n\n## Busting cache for a specific function call\n\nYou can force expire (bust) the cache for a specific function call\n\n```doctest\n>> await addition(3, 4)\n>> await addition(3, 7)\n>> await addition.bust(3, 4)\n```\n\nIn Redis you would see one entry as the other one has been busted\n\n```shell\n127.0.0.1:6379> KEYS *\n1) "__main__:addition:bb6b7afb6a6cf3191f6d7fd35d976d42"\n```\n\n\n## Busting cache for all function calls of a specific function\n\nYou can bust the cache for all instance of a function call\n\n```doctest\n>> await addition(3, 4)\n>> await addition(3, 7)\n>> await multiplication(2, 4)\n>> await addition.bust_all()\n```\n\nIn Redis you would see no entries for the function `addition` which has been busted,\nyou would see one entry for `multiplication`\n\n```shell\n127.0.0.1:6379> KEYS *\n1) "__main__:multiplication:bc3b7afc6a7cf3191f6d1fd31d810d55"\n```\n\n\n## Busting cache for all function calls of all functions\n\nYou can bust the cache of all entries made by Cache Money as well\n\n```doctest\n>> await addition(3, 4)\n>> await addition(3, 7)\n>> await multiplication(2, 4)\n>> cache_money.bust()\n```\n\nIn Redis you would see no entries\n\n```shell\n127.0.0.1:6379> KEYS *\n(empty array)\n```\n\n\n## Contributing and getting set up for local development\n\nTo set yourself up for development on Cache Money, make sure you are using\n[poetry](https://poetry.eustace.io/docs/) and simply run the following commands from the root directory:\n\n```shell\nmake sys-deps\nmake install\n```\n',
    'author': 'Alexandre Jutras',
    'author_email': 'alexandre.jutras@novisto.com',
    'maintainer': 'Alexandre Jutras',
    'maintainer_email': 'alexandre.jutras@novisto.com',
    'url': 'https://github.com/novisto/cache-money',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import inspect
from typing import Callable


def decode(s):
    return s.decode("utf-8") if isinstance(s, bytes) else s


def get_function_path(fn: Callable) -> str:
    return f"{inspect.getmodule(fn).__name__}.{fn.__name__}"
